import unittest
from thornpy.mechanics import contact

class Test_Contact(unittest.TestCase):

    def setUp(self):
        return

    def test_spherical(self):
        k, e = contact.spherical(2, None, 30000000, 30000000, .3, .3)
        self.assertAlmostEqual(k, 31081616.7554, 3)

    def tearDown(self):
        return




"""Miscellaneous tools.

"""
import csv
import os
import re
from numbers import Number
from pathlib import Path
from subprocess import Popen
from typing import Union

import pandas as pd

from scipy.io import loadmat


def open_in_explorer(path: os.PathLike) -> None:
    if Path(path).is_dir():
        _ = Popen(f'explorer.exe /root,"{path}"')
    elif Path(path).is_file():
        _ = Popen(f'explorer.exe /select,"{path}"')
    else:
        raise FileNotFoundError()


def num_to_ith(num):
    """Converts an integer to a string containing an ordinal number (1st, 2nd, 3rd, ect.)

    Parameters
    ----------
    num : int
        Number

    Returns
    -------
    str
        Ordinal number

    """
    if num == -1:
        return 'last'
    elif num < -1:
        value = str(num+1).replace('-', '')
    else:
        value = str(num)

    last_digit = value[-1]

    if len(value) > 1 and value[-2] == '1':
        suffix = 'th'
    elif last_digit == '1':
        suffix = 'st'
    elif last_digit == '2':
        suffix = 'nd'
    elif last_digit == '3':
        suffix = 'rd'
    else:
        suffix = 'th'

    if num < -1:
        suffix += ' to last'

    return value + suffix


def num_to_str(num: Number,
               places: int,
               allow_less=False,
               use_scientific: Union[bool, float] = False) -> str:
    """Converts a number to a string with a given number of significant figures.

    Parameters
    ----------
    num : Number
        Number to be converted
    places : int
        Number of decimal places to use
    allow_less : bool, optional
        If True, allows the number to be converted to a string with less decimal places than 
        `places`, by default False
    use_scientific : bool or Number, optional
        If True, uses scientific notation. If False, uses decimal notation. 
        If a number is passed, uses scientific notation if `num` is greater 
        than `use_scientific`, by default False

    Returns
    -------
    str
        Number converted to a string with a given number of decimal places

    """
    if not isinstance(use_scientific, bool) and isinstance(use_scientific, Number):
        if use_scientific > 1:
            use_scientific = abs(num) >= abs(use_scientific)
        else:
            use_scientific = abs(num) <= abs(use_scientific)

    if use_scientific:
        string = f'{num:.{places}e}'
    else:
        string = f'{num:.{places}f}'

    if allow_less and '.' in string:

        left, right, *exp = re.split(r'\.|e', string, flags=re.IGNORECASE)
        right = str(round(float(f'0.{right}'), places)).replace('0.', '.')
        if right == '.0':
            right = ''
        string = 'e'.join([f'{left}{right}', *exp])

    return string


def read_data_string(text, delimiter=',', newline='\n', has_headerline=True):
    """Reads a delimited string into a list of dictionaries.  Functions very similar to :meth:`numpy.genfromtxt`, but for strings instead of text files.

    Parameters
    ----------
    text : str
        String of row/column data with delimiters and new line indicators given in `delimiter` and `newline`.
    delimiter : str, optional
        Delimiter used in `text`, by default ','
    newline : str, optional
        New line indicator used in `text`, by default '\n'
    has_headerline : bool, optional
        If True, treats the first line of `text` as headers. If False, treats the first line of `text` as data and makes generic headers, by default True

    Returns
    -------
    :obj:`list` of :obj:`dict`
        A list of dictionaries containing the data from `text`

    """
    lines = text.split(newline)

    # Generate headers
    if has_headerline:
        # If the text has headerlines, get them
        headers = lines.pop(0).split(delimiter)

    else:
        # If the text doesn't have headerlines, make generic ones
        headers = [str(i+1) for i in range(len(lines[0].split(delimiter)))]

    data = []
    for line in lines:
        # For each line, check if data is missing

        if len(line.split(delimiter)) == len(headers):
            # If there is no missing data on this line, initialize a dictionary for the line data
            line_data = {}

            for header, value in zip(headers, line.split(delimiter)):
                # For each column in the line, add to the line_data dict (header as key and value as value)
                line_data[header] = value

            # Append the line_data dict to the data list
            data.append(line_data)

    return data


def convert_path(filepath):
    """Converts the slashes in `filepath` to whatever is appropriate for the current OS.

    Parameters
    ----------
    filepath : str
        Filepath to be converted

    Returns
    -------
    str
        Converted filepath

    """
    return os.path.normpath(filepath.replace('\\', os.sep).replace('/', os.sep))


def mat_to_pd(filename, data_start=None, data_end=None):
    """Creates a pandas dataframe from a .mat file

    Parameters
    ----------
    filename : str
        Filename of a .mat file

    Returns
    -------
    dataframe
        Pandas datafrom of data in `filename`

    """
    mat = loadmat(filename)  # load mat-file

    dataframes = []

    for variable in [var for var in mat if var not in ['__header__', '__globals__', '__version__']]:

        if mat[variable].shape[1] == 1:
            array = mat[variable]
        elif mat[variable].shape[0] == 1:
            array = mat[variable].transpose()
        else:
            raise ValueError(f'{filename} does not contain the expected data!')

        # Figure out data start and end indices if given
        if data_start is None:
            data_start = 0
        if data_end is None:
            data_end = len(array)

        dataframes.append(pd.DataFrame(array[data_start:data_end], columns=[variable]))

    dataframe = pd.concat(dataframes, axis=1)

    return dataframe


def dict_to_csv(data, filename):
    """Writes data in `dict` to a csv file with the keys as headers and the values as columns.

    Parameters
    ----------
    data : dict
        Dictionary of data
    filename : str
        Name of file to write data to. (include extension)

    """
    with open(filename, 'w', newline='') as outfile:
        writer = csv.writer(outfile)
        writer.writerow(data.keys())
        writer.writerows(zip(*data.values()))


def display_df_side_by_side(df1: pd.DataFrame, df2: pd.DataFrame):
    """Displays two `pd.DataFrames` side by side in a jupyter notebook

    Note
    ----
    Only works in a jupyter notebook

    Parameters
    ----------
    df1 : pd.DataFrame
        Left `pd.DataFrame`
    df2 : pd.DataFrame
        Right `pd.DataFrame`
    """
    from IPython.display import display_html
    df1_styler = df1.style.set_table_attributes("style='display:inline'").set_caption('df1')
    df2_styler = df2.style.set_table_attributes("style='display:inline'").set_caption('df2')

    display_html(df1_styler._repr_html_() + df2_styler._repr_html_(), raw=True)



import io
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.figure import Figure
from PIL import Image


def plot_3d(x, y, z, figure=None, x_label='X', y_label='Y', z_label='Z'):
    """Creates a 3D plot using matplotlib

    Parameters
    ----------
    x : list
        X values (can be :class:`numpy.array` or :class:`Pandas.Series`)
    y : list
        Y values (can be :class:`numpy.array` or :class:`Pandas.Series`)
    z : list
        Z values (can be :class:`numpy.array` or :class:`Pandas.Series`)
    figure : matplotlib.figure.Figure, optional
        Existing figure to add to (the default is None, which creates a new figure)
    x_label : str, optional
        X axis label (the default is 'X')
    y_label : str, optional
        Y axis label (the default is 'Y')
    z_label : str, optional
        Z axis label (the default is 'Z')

    Returns
    -------
    matplotlib.figure.Figure
        3D plot of x, y, and z

    """
    if figure is None:
        figure = plt.figure()
    axis = figure.gca(projection='3d')
    axis.plot(x, y, z, zdir='z')
    axis.set_xlabel(x_label)
    axis.set_ylabel(y_label)
    axis.set_zlabel(z_label)
    axis.axis('equal')

    return figure


def heatmap(x, y, **kwargs):
    """From https://github.com/drazenz/heatmap

    """
    if 'color' in kwargs:
        color = kwargs['color']
    else:
        color = [1]*len(x)

    if 'palette' in kwargs:
        palette = kwargs['palette']
        n_colors = len(palette)
    else:
        n_colors = 256  # Use 256 colors for the diverging color palette
        palette = sns.color_palette("Blues", n_colors)

    if 'color_range' in kwargs:
        color_min, color_max = kwargs['color_range']
    else:
        color_min, color_max = min(color), max(color)  # Range of values that will be mapped to the palette, i.e. min and max possible correlation

    def value_to_color(val):
        if color_min == color_max:
            return palette[-1]
        else:
            # position of value in the input range, relative to the length of the input range
            val_position = float((val - color_min)) / (color_max - color_min)
            val_position = min(max(val_position, 0), 1)  # bound the position betwen 0 and 1
            ind = int(val_position * (n_colors - 1))  # target index in the color palette
            return palette[ind]

    if 'size' in kwargs:
        size = kwargs['size']
    else:
        size = [1]*len(x)

    if 'size_range' in kwargs:
        size_min, size_max = kwargs['size_range'][0], kwargs['size_range'][1]
    else:
        size_min, size_max = min(size), max(size)

    size_scale = kwargs.get('size_scale', 500)

    def value_to_size(val):
        if size_min == size_max:
            return 1 * size_scale
        else:
            # position of value in the input range, relative to the length of the input range
            val_position = (val - size_min) * 0.99 / (size_max - size_min) + 0.01
            val_position = min(max(val_position, 0), 1)  # bound the position betwen 0 and 1
            return val_position * size_scale
    if 'x_order' in kwargs:
        x_names = [t for t in kwargs['x_order']]
    else:
        x_names = [t for t in sorted(set([v for v in x]))]
    x_to_num = {p[1]: p[0] for p in enumerate(x_names)}

    if 'y_order' in kwargs:
        y_names = [t for t in kwargs['y_order']]
    else:
        y_names = [t for t in sorted(set([v for v in y]))]
    y_to_num = {p[1]: p[0] for p in enumerate(y_names)}

    plot_grid = plt.GridSpec(1, 15, hspace=0.2, wspace=0.1)  # Setup a 1x10 grid
    ax = plt.subplot(plot_grid[:, :-1])  # Use the left 14/15ths of the grid for the main plot

    marker = kwargs.get('marker', 's')

    kwargs_pass_on = {k: v for k, v in kwargs.items() if k not in [
        'color', 'palette', 'color_range', 'size', 'size_range', 'size_scale', 'marker', 'x_order', 'y_order'
    ]}

    ax.scatter(
        x=[x_to_num[v] for v in x],
        y=[y_to_num[v] for v in y],
        marker=marker,
        s=[value_to_size(v) for v in size],
        c=[value_to_color(v) for v in color],
        **kwargs_pass_on
    )
    ax.set_xticks([v for k, v in x_to_num.items()])
    ax.set_xticklabels([k for k in x_to_num], rotation=45, horizontalalignment='right')
    ax.set_yticks([v for k, v in y_to_num.items()])
    ax.set_yticklabels([k for k in y_to_num])

    ax.grid(False, 'major')
    ax.grid(True, 'minor')
    ax.set_xticks([t + 0.5 for t in ax.get_xticks()], minor=True)
    ax.set_yticks([t + 0.5 for t in ax.get_yticks()], minor=True)

    ax.set_xlim([-0.5, max([v for v in x_to_num.values()]) + 0.5])
    ax.set_ylim([-0.5, max([v for v in y_to_num.values()]) + 0.5])
    ax.set_facecolor('#F1F1F1')

    # Add color legend on the right side of the plot
    if color_min < color_max:
        ax = plt.subplot(plot_grid[:, -1])  # Use the rightmost column of the plot

        col_x = [0]*len(palette)  # Fixed x coordinate for the bars
        bar_y = np.linspace(color_min, color_max, n_colors)  # y coordinates for each of the n_colors bars

        bar_height = bar_y[1] - bar_y[0]
        ax.barh(
            y=bar_y,
            width=[5]*len(palette),  # Make bars 5 units wide
            left=col_x,  # Make bars start at 0
            height=bar_height,
            color=palette,
            linewidth=0
        )
        ax.set_xlim(1, 2)  # Bars are going from 0 to 5, so lets crop the plot somewhere in the middle
        ax.grid(False)  # Hide grid
        ax.set_facecolor('white')  # Make background white
        ax.set_xticks([])  # Remove horizontal ticks
        ax.set_yticks(np.linspace(min(bar_y), max(bar_y), 3))  # Show vertical ticks for min, middle and max
        ax.yaxis.tick_right()  # Show vertical ticks on the right


def corrplot(data, size_scale=500, marker='s'):
    """From https://github.com/drazenz/heatmap

    """
    corr = pd.melt(data.reset_index(), id_vars='index')
    corr.columns = ['x', 'y', 'value']
    corr.fillna(0, inplace=True)
    heatmap(
        corr['x'], corr['y'],
        color=corr['value'], color_range=[-1, 1],
        palette=sns.diverging_palette(20, 220, n=256),
        size=corr['value'].abs(), size_range=[0, 1],
        marker=marker,
        x_order=data.columns,
        y_order=data.columns[::-1],
        size_scale=size_scale
    )


def save_multifig(figs: List[Figure], filename: Path):
    """Save multiple figures to a single file."""
    imgs = [io.BytesIO() for _ in figs]
    for fig, img in zip(figs, imgs):
        fig.savefig(img, format="png")
    imgs = [Image.open(img) for img in imgs]
    imgs_comb = np.hstack((np.asarray(i) for i in imgs))
    imgs_comb = Image.fromarray(imgs_comb)
    imgs_comb.save(filename)
    return imgs_comb


def concat_images(image_files: List[Path], filename: Path):
    """Concatenate images horizontally."""
    imgs = [Image.open(img) for img in image_files]
    imgs_comb = np.hstack((np.asarray(i) for i in imgs))
    imgs_comb = Image.fromarray(imgs_comb)
    imgs_comb.save(filename)


def _save_multifigs(figs, filename: Path):
    filename = Path(filename)
    image_files = []

    with TemporaryDirectory() as tmpdir:
        for i, fig in enumerate(figs):
            image_file = Path(tmpdir) / f'fig_{i}{filename.suffix}'
            fig.savefig(image_file)
            image_files.append(image_file)

        _concat_images(image_files, filename)


def _concat_images(input_files, output_file, vertical=True):
    images = [Image.open(x) for x in input_files]
    widths, heights = zip(*(i.size for i in images))

    max_width = max(widths)
    total_height = sum(heights)

    new_im = Image.new('RGB', (max_width, total_height))

    offset = 0
    for im in images:
        new_im.paste(im, (0, offset) if vertical else (offset, 0))
        offset += im.size[1 if vertical else 0]

    new_im.save(output_file)



"""Module for contact mechanics solutions.

"""
#pylint: disable=invalid-name

def spherical(r_1, r_2, E_1, E_2, vu_1, vu_2, P_guess=None):    
    """Returns the contact parameters for two spheres in contact or for a sphere in contact with a flat surface.  In the case of a flat surface, enter None for the radius.
    
    Parameters
    ----------
    r_1 : {None, float}
        Radius of sphere 1
    r_2 : float or None
        Radius of sphere 2
    E_1 : float
        Modulus of sphere 1
    E_2 : float
        Modulus of sphere 2
    vu_1 : float
        Poisson's ratio for sphere 1
    vu_2 : float
        Poisson's ratio for sphere 2
    P_guess : float, optional
        Estimate of the applied load, by default None
    
    Returns
    -------
    float
        stiffness
    float
        stiffness exponent
    float
        contact area (Only if P_guess is given)
    float 
        max penetration parameter for use in an MSC Adams model. (Only if P_guess is given)

    Raises
    ------
    ValueError
        Raised if both radiuses are none

    """    
    # Effective radius
    if r_1 is None:
        R = r_2
    elif r_2 is None:
        R = r_1
    elif r_1 is None and r_2 is None:
        raise ValueError('At least one radius must be provided!')
    else:        
        R = (r_1**-1 + r_2**-1)**-1

    # Effective Modulus
    E = ((1-vu_1**2)/E_1 + (1-vu_2**2)/E_2)**-1

    # Contact Stiffness
    k = 4/3*E*R**.5

    # Stiffness Exponent
    e = 3/2

    if P_guess is not None:
        d = (9*P_guess**2/(16*r_1*E**2))**(1/3)
        d_max = 0.3*d
        a = (3*P_guess*R/4/E)**(1/3)        
        return k, e, a, d_max

    else:
        return k, e



"""Signal processing module
"""
from scipy.signal import butter, filtfilt, find_peaks
from scipy.signal.windows import hann, hamming, blackman, boxcar
import matplotlib.pyplot as plt
from matplotlib import cm  
from matplotlib.colors import Normalize
from matplotlib.lines import Line2D
import numpy as np
import pandas as pd

def autocorr(x):
    x = np.array(x)
    n = x.size
    norm = (x - np.mean(x))
    result = np.correlate(norm, norm, mode='same')
    acorr = result[n//2 + 1:] / (x.var() * np.arange(n-1, n//2, -1))
    lag = np.abs(acorr).argmax() + 1
    r = acorr[lag-1]        
    return r, lag

def low_pass(sig, time, freq_cutoff, N=5):
    """Applies a Nth order butterworth lowpass filter.
    
    Parameters
    ----------
    sig : list
        Signal to be filtered
    time : list
        Time signal in seconds associated with `sig`
    freq_cutoff : int
        Cutoff frequency in Hz to use when filtering
    N : int
        Order of butterworth filter (default is 5)
    
    Returns
    -------
    list
        Filtered signal
    list
        Time signal associated with filtered signal

    """
    freq_samp = (len(time)-1)/(time[-1]-time[0])
    omega = freq_cutoff/freq_samp*2
    b_coef, a_coef = butter(N, omega)
    return list(filtfilt(b_coef, a_coef, sig)), time
        
def step_function(index, start, init_val, end, final_val):
    """Approximates the Heaviside step function with a cubic polynomial.
    
    Example
    -------
    >>> x = [2, 3, 3.5, 4, 5]
    >>> start = 3
    >>> init_val = 0
    >>> end = 3
    >>> final_val = 1
    >>> step_function(x, start, init_val, end, final_val)
    [0.0, 0.0, 0.5, 1.0, 1.0]
    
    Parameters
    ----------
    index : list
        Independent variable
    start : float
        Value of independent variable at which the STEP function begins
    init_val : float
        Initial value of the step
    end : float
        Value of independent variable at which the STEP function ends
    final_val : float
        Final value of the step
    
    Returns
    -------
    list
        Resulting step function

    """
    height = final_val - init_val
    step =  [init_val for ind in index if ind <= start]    
    step += [init_val + height*((ind-start)/(end-start))**2 * (3-2*((ind-start)/(end-start))) for ind in index if start < ind < end]
    step += [final_val for ind in index if ind >= end]
    return step

def fft_watefall(time, sig, percent_overlap=50, n_fft=1024, title=None, t_min=None, t_max=None, input_sig=None, input_conversion_factor=60/360, input_unit='RPM', response_unit=None, response_conversion_factor=1, psd=False, z_scale='linear', order_lines=None, f_range=None, clean_sig=None, return_order_cuts=None, vmin=None, vmax=None):
    """Genenerates a waterfall plot from data in an Adams result or request file.
    
    Parameters
    ----------
    time : list
        Time signal in seconds
    sig : list
        Signal
    comp : str
        Name of the result component in the Adams dataset
    percent_overlap : int, optional
        Percent overlap for the FFT waterfall, by default 50
    n_fft : int, optional
        Number of points used in each FFT, by default 1024
    window : str, optional
        Type of window used for each FFT, by default 'hanning'
    t_min : float, optional
        Start time if cropping data, by default None
    t_max : float, optional
        End time if cropping data, by default None
    vmin : float, optional
        Colorbar min (default is None which autoscales)
    vmax : float, optional
        Colorbar max (default is None which autoscales)    
    input_sig : list, optional
        Time signal of input shaft speed, by default None
    input_time : list, optional
        Input shaft speed, by default None
    input_conversion_factor : float, optional
        Conversion factor applied to input signal, by default 60/360 (deg/s to RPM)
    response_unit : str, optional
        Text to display on the response signal y axis, by default None
    input_unit : str, optional
        Text to display on the input signal y axis, by default None
    conversion_factor : float, optional
        Conversion factor applied to response signal, by default 1
    psd : bool, optional
        If True FFT will output in PSD. If False FFT will output Magnitude , by default False
    z_scale : str, optional
        The scaling of the values in the spec. 'linear' is no scaling. 'dB' returns the values in dB scale. `psd` is True, this is dB power (10 * log10). Otherwise this is dB amplitude (20 * log10). by default 'linear'
    orders : list, optional
        Adds order fans to the waterfall plots at each order in the list.  Only used if `input_comp` is given.
    f_range : list, optional
        [mininum frequency, maximum frequency].  Only affects plot limits
    clean_sig : float, optional
        If given, removes data points that exceed `clean_sig` multiplied by the standard deviation of the signal.
    title : str, optional
        Plot title, by default None.
    
    Returns
    -------
    Figure
        Waterfall plot

    """    
    # Convert to other unis (for converting to Gs)
    if response_conversion_factor is not None:
        sig = [v*response_conversion_factor for v in sig]
        
    if clean_sig is not None:
        sig, i_removed, dirty_sig = _clean_sig(sig, clean_sig)
    
    check_num_points(len(time), n_fft)

    t_s = (time[-1] - time[0])/(len(time) - 1)
    f_s = 1/t_s
    
    _fig, (ax1, ax2) = plt.subplots(nrows=2)
    ax1.plot(time, sig)

    Pxx, freqs, bins, im = ax2.specgram(sig, NFFT=n_fft, Fs=f_s, noverlap=percent_overlap/100*n_fft, detrend='mean', mode='magnitude' if psd is False else 'psd', scale=z_scale, cmap='nipy_spectral')
    # The `specgram` method returns 4 objects. They are:
    # - Pxx: the periodogram
    # - freqs: the frequency vector
    # - bins: the centers of the time bins
    # - im: the matplotlib.image.AxesImage instance representing the data in the plot
    _fig.clear()
    plt.close()    
    del _fig

    # Convert to dB
    if z_scale.lower() == 'db':
        Pxx = np.log10(Pxx) * (20 if psd is False else 10)

    # ----------------
    # Signal
    # ----------------
    x_wtrfl, y_wtrfl = np.meshgrid(bins+time[0], freqs)
    fig, axes = plt.subplots(nrows=2 if input_sig is None else 3)
        
    axes[0].plot(time, sig, zorder=2)

    y_lim = axes[0].get_ylim()
    
    # if clean_sig is not None:
    #     axes[0].plot(time, dirty_sig, '--', linewidth=1, zorder=1)
    #     axes[0].plot([time[i] for i in i_removed], [sig[i] for i in i_removed], '.', markersize=3, zorder=3)

    axes[0].set_xlim(time[0], time[-1])
    axes[0].set_ylim(y_lim[0], y_lim[1])

    if response_unit is None:
        axes[0].set_ylabel(response_unit)
    else:
        axes[0].set_ylabel(response_unit)
    axes[0].grid()        

    if title is not None:
        axes[0].set_title(title)    

    if input_sig is None:
        # ----------------
        # 3D FFT
        # ----------------        
        if vmin is None:
            vmin = Pxx.min()
        if vmax is None:
            vmax = Pxx.max()
        cmap = cm.get_cmap('nipy_spectral', 100)
        wtrfl_surf = axes[1].contourf(x_wtrfl, y_wtrfl, Pxx, 250, cmap=cmap, extend='both', vmin=vmin, vmax=vmax)
        axes[1].set_xlim(time[0], time[-1])
        axes[1].set_xlabel('Time (s)')
        axes[1].set_ylabel('Frequency (Hz)')
        axes[1].set_grid(True)
        # wtrfl_fig.colorbar(wtrfl_surf)

    else:        
        # ----------------
        # Order Chart
        # ----------------
        if len(input_sig) != len(time):
            raise ValueError('Input signal must be the same lenght as the response signel.')

        input_sig_rpm = [abs(v*input_conversion_factor) for v in input_sig]
        input_bins = [abs(input_sig_rpm[np.argmax(time>=t_bn+time[0])]) for t_bn in bins]
        x_order, y_order = np.meshgrid(input_bins, freqs)
        
        # Plot Input Signal
        axes[1].plot(time, input_sig_rpm)
        axes[1].set_xlim(time[0], time[-1])
        axes[1].set_xlabel('Time (s)')
        axes[1].set_ylabel(input_unit)
        axes[1].set_title('Input Speed')
        axes[1].grid(True)

        # Plot Order Waterfall
        if f_range is not None:
            i_min = int(np.argmax(freqs>=f_range[0])) if min(freqs)<f_range[0] else 0
            i_max = int(np.argmax(freqs>=f_range[1])) if max(freqs)>f_range[1] else len(freqs)
        else:
            i_min = 0
            i_max = len(freqs)
 
        if vmin is None:
            vmin = Pxx[i_min:i_max].min()
        if vmax is None:
            vmax = Pxx[i_min:i_max].max()
        cmap = cm.get_cmap('nipy_spectral', 100)
        wtrfl_surf = axes[2].contourf(x_order[i_min:i_max], y_order[i_min:i_max], Pxx[i_min:i_max], 250, cmap=cmap, extend='both', vmin=vmin, vmax=vmax)
        axes[2].set_xlabel(input_unit)
        axes[2].set_ylabel('Frequency (Hz)')

        # Add order lines
        if order_lines is not None:
            _add_order_lines(axes[2], order_lines, input_unit)
    
        if psd is True:
            y_label = f'{response_unit}^2/Hz' if z_scale.lower()!='db' else 'dB'            
        elif response_unit is not None:
            y_label = response_unit if z_scale.lower()!='db' else 'dB'

        # Add colorbar
        sm = cm.ScalarMappable(norm=Normalize(vmin, vmax), cmap=cmap)
        sm.set_array([vmin, vmax])
        sm.autoscale()
        cbar = fig.colorbar(sm, ticks=np.linspace(vmin, vmax, 10))
        cbar.set_label(y_label)

        # Generate Order Cut Plots
        if return_order_cuts is not None:
            order_cuts = _order_cut_plot(input_bins, freqs[i_min:i_max], Pxx[i_min:i_max], return_order_cuts, y_label=y_label)
        
    fig.set_size_inches(10,10)
    fig.tight_layout()

    print(f'dt: {n_fft*t_s}')

    if input_sig is None:
        return (fig, time, sig)
    elif return_order_cuts is None:
        return (fig, time, sig, input_sig_rpm)
    else:
        return (fig, time, sig, input_sig_rpm, order_cuts)

def check_num_points(num, n_fft):    
    if num < n_fft:
        for pow_2 in [2**e for e in range(100)][::-1]:
            if num >= pow_2:
                suggested_n = pow_2
                raise ValueError(f'n_fft ({n_fft}) must be less than the number of points in the signal ({num}). Try {suggested_n}.')        
        
        raise ValueError(f'n_fft ({n_fft}) must be less than the number of points in the signal ({num}).')

def get_window(window):
    if window == 'hanning':
        win = hann
    elif window == 'hamming':
        win = hamming
    elif window == 'blackman':
        win = blackman
    elif window in ['rectangular', 'boxcar', None]:
        win = boxcar
    else:
        raise ValueError(f'{window} not recognized as a window!')    
    return win

def _add_order_lines(ax, orders, input_unit):
    """Adds order fan lines to ax at each order in `orders`.
    
    Parameters
    ----------
    ax : axes
        An axes object with a order waterfall countour plot
    orders : list
        List of orders to put lines at.
    input_to_hz : float, optional
        Value to scale the x value by to convert to hz, by default 1/60 (rpm to hz)

    """
    input_to_hz = _get_conversion_to_hz(input_unit)
    ax.autoscale(False)

    for order in orders:        
        x_coords = [0, ax.get_xlim()[-1]]
        y_coords = [0, x_coords[-1]*input_to_hz*order]

        if y_coords[-1] > ax.get_ylim()[-1]:
            y_coords[-1] = ax.get_ylim()[-1]
            x_coords[-1] = y_coords[-1]/input_to_hz/order

        # Add line
        ax.plot(x_coords, y_coords, color='white', linestyle='-', linewidth=2, alpha=.75)

        # Add annotation
        box_props = dict(boxstyle='round', pad=.3, fc='white', ec='black')
        ax.annotate(f'{order:.0f}', (x_coords[-1], y_coords[-1]), bbox=box_props)

def _get_conversion_to_hz(input_unit):
    """Given a unit returns the factor to convert to Hz.
    
    Parameters
    ----------
    input_unit : str
        Unit to be converted to Hz.  Must be 'rpm', 'hz', 'deg/s', 'degs/s', 'deg/sec', or 'degs/sec'.
    
    Returns
    -------
    float
        conversion factor

    """
    # Get conversion factor
    if input_unit.lower() == 'rpm':
        input_to_hz = 1/60
    elif input_unit.lower() == 'hz':
        input_to_hz = 1
    elif input_unit.lower().startswith('deg') and (input_unit.lower().endswith('/s') or input_unit.lower().endswith('/sec')):
        input_to_hz = 1/360

    return input_to_hz

def _clean_sig(sig, n_sigma):
    cleaned_sig = np.array(sig)
    i_pks, _ = find_peaks(np.abs(cleaned_sig), prominence=np.std(cleaned_sig)*n_sigma)

    for i_pk in i_pks:
        cleaned_sig[i_pk] = np.mean([cleaned_sig[i_pk-1], cleaned_sig[i_pk+1]])

    return list(cleaned_sig), list(i_pks), list(sig)

def _order_cut_plot(input_sig, freqs, Pxx, orders, input_to_hz=1/60, input_unit='rpm', y_label=None):
    """Returns a figure with suplots of order cuts.
    
    Parameters
    ----------
    input_sig : list
        Input speed signal corresponding to `Pxx` bins
    freqs : list
        Frequencies crresponding to `Pxx` bins
    Pxx : 2-D array
        Spectrum
    orders : list
        Orders to plot
    input_to_hz : float, optional
        Conversion from input units to Hz, by default 1/60 (rpm to hz)
    input_unit : str, optional
        Input signal units, by default 'rpm'
    y_label : str, optional
        Y axis lable for plots, by default None
    
    Returns
    -------
    figure
        figure with suplots of order cuts

    """
    fig, axes = plt.subplots(nrows=len(orders))
    for ax, order in zip(axes, orders):

        # Get the speeds that keep this order within the range in freqs
        speeds_in_range = [speed for speed in input_sig if min(freqs) <= speed*input_to_hz*order <= max(freqs)]

        # Get the order frequencies
        order_freqs = [speed*input_to_hz*order for speed in speeds_in_range]

        i_freqs = [_find_nearest(freqs, f) for f in order_freqs]
        
        # Get the order amplitudes
        amp = [Pxx[i][j] for i,j in zip(i_freqs, range(len(speeds_in_range)))]
        
        ax.plot(speeds_in_range, amp)
        ax.set_xlabel(input_unit)

        if y_label is not None:
            ax.set_ylabel(y_label)

        ax.set_title(f'Order {order}')
        ax.grid()

    fig.set_size_inches(10,10/3*len(orders))
    fig.tight_layout()

    return fig

def _find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx

def remove_data_point(x, y, i):
    """Removes the data point at index `i` and replaces it with a linear interpolation between the neighboring points.

    Parameters
    ----------
    x : list
        X coordinates of data
    y : list
        Y coordinates of data
    i : int
        Index of data point to remove

    Returns
    -------
    list
        Y coordinates of data with point removed
        
    """
    # Validate data
    if len(y) < 3:
        raise ValueError(f'The length of the signal must be greater than 3.  The signal passed has length {len(y)}!')

    if i == 0:
        # If the first point is passed, extrapolate backwards
        y[i] = np.interp(x[i], [x[i+1], x[i+2]], [y[i+1], y[i+2]])
    elif i == len(y):
        # If the last point is passed, extrapolate forwards
        y[i] = np.interp(x[i], [x[i-2], x[i-1]], [y[i-2], y[i-1]])
    else:
        # If any other point is passed, interpolate
        y[i] = np.interp(x[i], [x[i-1], x[i+1]], [y[i-1], y[i+1]])

    return y

def manually_clean_sigs(x, y, print_debug=False, indices=False, _fig=None):
    """Opens a plot window allowing the user to manually select points to remove from a signal.

    Parameters
    ----------
    x : list
        X coordinates of data to clean
    y : list of lists
        Y coordinates of data to clean
    print_debug : bool, optional
        prints the modified data points, by default False
    indices : bool, optional
        If True, returns the modified indices, by default False

    """        
    def onpick(event, _fig, x, y):
        if isinstance(event.artist, Line2D) and event.ind[0] not in [0, len(x)-1]:
            ind = event.ind[0]            
            
            if print_debug is True:
                print(f'x={x[ind]}   |   i={ind}   |   lenx={len(x)}')
            #     print(f'({x[ind]}, {y[ind]})   |   i={ind}   |   lenx={len(x)}')
            
            for each_y in y:
                each_y = remove_data_point(x, each_y, ind)
            
            # Append the index to the modified indices list
            i_mod.append(ind)

            _fig.clear()
            manually_clean_sigs(x, y, _fig=_fig)
    
    x = list(x)
    y = [list(vals) for vals in y]

    # Initialize a list of modified indices
    i_mod = []

    if _fig is None:
        _fig, ax = plt.subplots()
        show_new = True
    else:
        _fig.clear()
        ax = _fig.add_subplot(1,1,1)
        show_new = False

    for each_y in y:
        ax.plot(x, each_y, linestyle='-', linewidth='1', marker='.', markersize=5, picker=2)
    
    _fig.canvas.mpl_connect('pick_event', lambda evt: onpick(evt, _fig, x, y))

    ax.grid()  
    
    if show_new is True:
        plt.show()
        plt.close(_fig)
    else:
        plt.draw()

    if indices is True:
        return y, i_mod
    else:
        return y


def manually_clean_sig(x, y, print_debug=False, indices=False, _fig=None):
    """Opens a plot window allowing the user to manually select points to remove from a signal.

    Parameters
    ----------
    x : list
        X coordinates of data to clean
    y : list
        Y coordinates of data to clean
    print_debug : bool, optional
        prints the modified data points, by default False
    indices : bool, optional
        If True, returns the modified indices, by default False

    """        
    def onpick(event, _fig, x, y):
        if isinstance(event.artist, Line2D) and event.ind[0] not in [0, len(x)-1]:
            ind = event.ind[0]            
            
            if print_debug is True:
                print(f'({x[ind]}, {y[ind]})   |   i={ind}   |   lenx={len(x)}')
            
            y = remove_data_point(x, y, ind)
            
            # Append the index to the modified indices list
            i_mod.append(ind)

            _fig.clear()
            manually_clean_sig(x, y, _fig=_fig)
    
    x = list(x)
    y = list(y)

    # Initialize a list of modified indices
    i_mod = []

    if _fig is None:
        _fig, ax = plt.subplots()
        show_new = True
    else:
        _fig.clear()
        ax = _fig.add_subplot(1,1,1)
        show_new = False

    ax.plot(x, y, linestyle='-', linewidth='1', marker='.', markersize=5, picker=2)
    ax.grid()  

    _fig.canvas.mpl_connect('pick_event', lambda evt: onpick(evt, _fig, x, y))
    
    if show_new is True:
        plt.show()
        plt.close(_fig)
    else:
        plt.draw()

    if indices is True:
        return y, i_mod
    else:
        return y

def resample_dataframe_non_time(
        df: pd.DataFrame,
        n_samples: int,
        column: str,
        x_range=None,
        apply_rolling_average=False,
        **rolling_args,
    ) -> pd.DataFrame:

    if x_range is None:
        x = np.linspace(df[column].min(), df[column].max(), n_samples)    
    else:
        x = np.linspace(x_range[0], x_range[1], n_samples)
        
    _df = (
        df.drop_duplicates(subset=column)
        .set_index(column, drop=True)
    )
    
    if apply_rolling_average:
        _df = _df.rolling(**rolling_args)
    
    df = (
        _df.reindex(_df.index.union(x))
        .interpolate('values')
        .loc[x]
        .reset_index(drop=False)
    )

    return df

def resample_dataframe(
        df: pd.DataFrame,
        resample_time: int,
        time_column='index',
        time_unit='s',
        datetime_column=None
    ) -> pd.DataFrame:
    """Resamples `df` using sample time `resample_time` in microseconds.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to resample
    resample_time : int
        Resample time in microseconds
    time_column : str, optional
        Name of time column, by default 'index'
    time_unit : str, optional
        Unit of time colujmn, by default 's'
    datetime_column : str, optional
        Name of datetime column, by default None

    Returns
    -------
    pd.DataFrame
        Resampled DataFrame
    """
    def resample(df: pd.DataFrame, datetime_column: str, resample_time: int) -> pd.DataFrame:
        """Resamples df
        
        Parameters
        ----------
        df : DataFrame
            Data Frame to be resampled
        resample_time : int
            New sample time in microseconds
        datetime_column : str
            Name of DateTime column
        """
        df.drop_duplicates(subset=datetime_column, inplace=True)
        
        # Set the index as the date (preserver current)
        index_name = df.index.name
        if index_name is not None:
            df = df.reset_index()
        df = df.set_index(datetime_column, drop=True)

        # Resample
        _df = df.resample(f'{resample_time}U').asfreq()

        # Join with original
        df = _df.join(df, lsuffix='__', how='outer')

        for column in [c for c in df.columns if c.endswith('__')]:
            # _tmpdf[column].fillna(_tmpdf[column[:-1]], inplace=True)
            del df[column]

        # df[datetime_column] = df.index.values
        
        df = df.apply(pd.Series.interpolate, args=('time',)).bfill()
        df = df.resample(f'{resample_time}U').asfreq()        

        df = df.reset_index()

        if index_name is not None:
            df = df.set_index(index_name, drop=True)
        
        return df

    if datetime_column is None and time_column == 'index':

        # If the time column is the index
        datetime_column = '__datetime'
        df[datetime_column] = pd.to_datetime(df.index, unit=time_unit) 
        df = resample(df, datetime_column, resample_time=resample_time)
        df = df.drop(datetime_column, axis=1)

    elif datetime_column is None:

        # If a time column is passed
        datetime_column = '__datetime'
        df[datetime_column] = pd.to_datetime(df[time_column], unit=time_unit)    
        df = resample(df, datetime_column, resample_time=resample_time)
        df = df.drop(datetime_column, axis=1)

    else:

        #  If a datetime column is passed
        df = resample(df, datetime_column, resample_time=resample_time)
    
    return df



from scipy.spatial.transform import Rotation as R

def euler_to_matrix(euler, sequence='ZXZ', degrees=False):
    """Convert Euler angles to rotation matrix.

    Parameters
    ----------
    euler : array_like
        Euler angles in radians.
    sequence : str, optional
        Euler sequence. Use capital to indicate extrinsic rotations. Default is 'ZXZ'.

    Returns
    -------
    matrix : ndarray
        Rotation matrix.

    """
    rot: R = R.from_euler(sequence, euler, degrees=degrees)
    return rot.as_matrix()



"""
Tools for generating PowerPoint presentations

"""
import os
import cv2

def get_movie_frame(movie_file, frame=0):
    """Creates a picture file of a specified frame from a movie file. Returns the filename of the picture file and the resolution of the picture.

    Parameters
    ----------
    movie_file : str
        Filename of a movie
    frame : int, optional
        Frame number to get, by default 0

    Returns
    -------
    str
        Filename of the picture file
    int
        Height of the picture in pixels
    int
        Width of the picture in pixels
        
    """
    movie = cv2.VideoCapture(movie_file)
    _, image = movie.read()    
    height, width, _ = image.shape
    filename = os.path.splitext(movie_file)[0] + f'_{frame}.jpg'
    cv2.imwrite(filename, image)
    
    return filename, height, width



"""Tests functions in the utilities module.

"""
import os
import unittest
from thornpy import utilities
from test import *


class Test_Utilities(unittest.TestCase):

    def setUp(self):
        return

    def test_num_to_ith(self):
        ordinals = [utilities.num_to_ith(num) for num in TEST_INTEGERS]
        self.assertEqual(ordinals, TEST_EXPECTED_ORDINALS)
    
    def test_read_data_string_with_headers(self):
        data = utilities.read_data_string(TEST_DATA_STRING)
        self.assertListEqual(data, TEST_EXPECTED_DATA_DICT_WITH_HEADERS)

    def test_read_data_string_without_headers(self):
        data = utilities.read_data_string(TEST_DATA_STRING, has_headerline=False)
        self.assertListEqual(data, TEST_EXPECTED_DATA_DICT_WITHOUT_HEADERS)
    
    def test_convert_path_windows(self):
        """Tests that utilities.convert_path works as expected.

        """
        filepath = 'home/thorn241/test'
        converted_filepath = utilities.convert_path(filepath)
        expected_converted_filepath = os.path.join('home', 'thorn241', 'test')
        self.assertEqual(converted_filepath, expected_converted_filepath)        

    def tearDown(self):
        return
        
class Test_NumToStr_PositiveLargeNumbers(unittest.TestCase):

    NUM = 1e2

    def test_zeros_after_decimal_when_not_allow_less(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=False)
        self.assertEqual(string, '100.000')
        
    def test_no_zeros_after_decimal_when_allow_less(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=False)
        self.assertEqual(string, '100')

    def test_zeros_after_decimal_when_not_allow_less_scientific(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=True)
        self.assertEqual(string, '1.000e+02')

        
    def test_no_zeros_after_decimal_when_allow_less_scientific(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=True)
        self.assertEqual(string, '1e+02')

    def test_zeros_after_decimal_when_not_allow_less_with_scientific_threshold_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=1e2)
        self.assertEqual(string, '1.000e+02')

        
    def test_no_zeros_after_decimal_when_allow_less_with_scientific_threshold_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=1e2)
        self.assertEqual(string, '1e+02')

    def test_zeros_after_decimal_when_not_allow_less_with_scientific_threshold_not_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=1e3)
        self.assertEqual(string, '100.000')

        
    def test_no_zeros_after_decimal_when_allow_less_with_scientific_threshold_not_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=1e3)
        self.assertEqual(string, '100')

        
class Test_NumToStr_NegativeLargeNumbers(unittest.TestCase):

    NUM = -1e2

    def test_zeros_after_decimal_when_not_allow_less(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=False)
        self.assertEqual(string, '-100.000')
        
    def test_no_zeros_after_decimal_when_allow_less(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=False)
        self.assertEqual(string, '-100')

    def test_zeros_after_decimal_when_not_allow_less_scientific(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=True)
        self.assertEqual(string, '-1.000e+02')

        
    def test_no_zeros_after_decimal_when_allow_less_scientific(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=True)
        self.assertEqual(string, '-1e+02')

    def test_zeros_after_decimal_when_not_allow_less_with_scientific_threshold_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=1e2)
        self.assertEqual(string, '-1.000e+02')

        
    def test_no_zeros_after_decimal_when_allow_less_with_scientific_threshold_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=1e2)
        self.assertEqual(string, '-1e+02')

    def test_zeros_after_decimal_when_not_allow_less_with_scientific_threshold_not_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=1e3)
        self.assertEqual(string, '-100.000')

        
    def test_no_zeros_after_decimal_when_allow_less_with_scientific_threshold_not_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=1e3)
        self.assertEqual(string, '-100')

        
class Test_NumToStr_PositiveSmallNumbers(unittest.TestCase):

    NUM = 1e-2

    def test_zeros_after_decimal_when_not_allow_less(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=False)
        self.assertEqual(string, '0.010')
        
    def test_no_zeros_after_decimal_when_allow_less(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=False)
        self.assertEqual(string, '0.01')

    def test_zeros_after_decimal_when_not_allow_less_scientific(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=True)
        self.assertEqual(string, '1.000e-02')

        
    def test_no_zeros_after_decimal_when_allow_less_scientific(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=True)
        self.assertEqual(string, '1e-02')

    def test_zeros_after_decimal_when_not_allow_less_with_scientific_threshold_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=1e-1)
        self.assertEqual(string, '1.000e-02')

        
    def test_no_zeros_after_decimal_when_allow_less_with_scientific_threshold_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=1e-1)
        self.assertEqual(string, '1e-02')

    def test_zeros_after_decimal_when_not_allow_less_with_scientific_threshold_not_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=1e-3)
        self.assertEqual(string, '0.010')

        
    def test_no_zeros_after_decimal_when_allow_less_with_scientific_threshold_not_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=1e-3)
        self.assertEqual(string, '0.01')

        
class Test_NumToStr_NegativeSmallNumbers(unittest.TestCase):

    NUM = -1e-2

    def test_zeros_after_decimal_when_not_allow_less(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=False)
        self.assertEqual(string, '-0.010')
        
    def test_no_zeros_after_decimal_when_allow_less(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=False)
        self.assertEqual(string, '-0.01')

    def test_zeros_after_decimal_when_not_allow_less_scientific(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=True)
        self.assertEqual(string, '-1.000e-02')

        
    def test_no_zeros_after_decimal_when_allow_less_scientific(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=True)
        self.assertEqual(string, '-1e-02')

    def test_zeros_after_decimal_when_not_allow_less_with_scientific_threshold_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=1e-1)
        self.assertEqual(string, '-1.000e-02')

        
    def test_no_zeros_after_decimal_when_allow_less_with_scientific_threshold_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=1e-1)
        self.assertEqual(string, '-1e-02')

    def test_zeros_after_decimal_when_not_allow_less_with_scientific_threshold_not_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=False, use_scientific=1e-3)
        self.assertEqual(string, '-0.010')

        
    def test_no_zeros_after_decimal_when_allow_less_with_scientific_threshold_not_exceeded(self):
        string = utilities.num_to_str(self.NUM, 3, allow_less=True, use_scientific=1e-3)
        self.assertEqual(string, '-0.01')



import os
import unittest

import matplotlib.pyplot as plt
from thornpy import geometry

class Test_GetArcArcPoints2(unittest.TestCase):

    def setUp(self):        
        return

    def test_get_arc_arc_points2(self):
        x_0 = -34
        y_0 = 28
        x_1 = -53
        y_1 = 56
        x_prev = -0
        y_prev = 0
        r_0 = 29
        end_angle = 0.02

        x_arc, y_arc, x_ca, y_ca, x_tan, y_tan, start_angle, i_center, x_cb, y_cb, r_b = geometry.get_arc_arc_points2(x_0, y_0, x_1, y_1, x_prev, y_prev, r_0, end_angle)

        plt.plot(x_arc, y_arc)
        plt.scatter(x_tan, y_tan, c='g')
        plt.scatter(x_prev, y_prev, c='k')
        plt.scatter(x_0, y_0, c='r')
        plt.scatter(x_1, y_1, c='r')
        plt.scatter(x_ca, y_ca, c='k', marker='+')
        plt.scatter(x_cb, y_cb, c='k', marker='+')
        plt.axis('square')
        plt.grid()

        plt.show()
        
        self.assertEqual(0,1)

    def tearDown(self):
        return


import os
from pickle import load, dump
import dill
import math
import sympy as sp
import numpy as np
from numpy.linalg import norm

from scipy import optimize

ARCLINE_FILE = os.path.join(os.path.dirname(__file__), 'arcline.pkl')
ARCARC_FILE = os.path.join(os.path.dirname(__file__), 'arcarc.pkl')
ARCARC2_FILE = os.path.join(os.path.dirname(__file__), 'arcarc2.pkl')

def get_arc_arc_points2(x_0, y_0, x_1, y_1, x_prev, y_prev, r_0, end_angle, pts=10):

    tan_end_ang = math.tan(end_angle)
    start_angle = get_line_ori(x_0, y_0, x_prev, y_prev)
    x_ca_multi = [x_0 + r_0*np.cos(start_angle + s*np.pi/2) for s in [1, -1]]
    y_ca_multi = [y_0 + r_0*np.sin(start_angle + s*np.pi/2) for s in [1, -1]]

    # Get center with largest y coordinate
    i_center = y_ca_multi.index(max(y_ca_multi))
    x_ca = x_ca_multi[i_center]
    y_ca = y_ca_multi[i_center]
      
    x_cb_sym = sp.Symbol('x_cb_sym', real=True)
    y_cb_sym = sp.Symbol('y_cb_sym', real=True)
    x_2_sym = sp.Symbol('x_2_sym', real=True)
    y_2_sym = sp.Symbol('y_2_sym', real=True)
    r_b_sym = sp.Symbol('r_b_sym', real=True)

    # Numerically
    eq_pt2_on_A = sp.Eq((x_2_sym - x_ca)**2 + (y_2_sym - y_ca)**2, r_0**2)
    eq_pt2_on_B = sp.Eq((x_2_sym - x_cb_sym)**2 + (y_2_sym - y_cb_sym)**2, r_b_sym**2)
    eq_pt3_on_B = sp.Eq((x_1 - x_cb_sym)**2 + (y_1 - y_cb_sym)**2, r_b_sym**2)
    eq_tangent_at_2 = sp.Eq((y_ca-y_2_sym) / (x_ca-x_2_sym), (y_2_sym-y_cb_sym) / (x_2_sym-x_cb_sym))
    eq_slope_at_3 = sp.Eq((x_1-x_cb_sym) / (y_1-y_cb_sym), tan_end_ang)

    init = (float(np.mean([x_0, x_1])), float(np.mean([y_0, y_1])), x_1, float(np.mean([y_0, y_1])), r_0)
    sols = sp.nsolve((eq_pt2_on_A, eq_pt2_on_B, eq_pt3_on_B, eq_tangent_at_2, eq_slope_at_3), (x_2_sym, y_2_sym, x_cb_sym, y_cb_sym, r_b_sym), init, verify=False)

    best_sol = _get_best_sol3(x_0, y_0, x_1, y_1, x_ca, y_ca, x_prev, y_prev, r_0, tan_end_ang, sols)
    x_tan, y_tan, x_cb, y_cb, r_1 = map(float, best_sol)
    
    # Figure out how many points in the two arcs (scale by length)
    arc1_length = r_0 * get_0_to_180(get_3point_ang((x_0, y_0), (x_ca, y_ca), (x_tan, y_tan)))
    arc2_length = r_1 * get_0_to_180(get_3point_ang((x_tan, y_tan), (x_cb, y_cb), (x_1, y_1)))
    arc_2_pts = round(pts/(arc1_length/arc2_length + 1))
    arc_1_pts = pts - arc_2_pts if arc_2_pts < pts else 1

    # Generate the points
    x_arc_1, y_arc_1 = get_arc_points(x_0, y_0, x_tan, y_tan, x_ca, y_ca, x_prev,      y_prev,      arc_1_pts, direction='cw')
    x_arc_2, y_arc_2 = get_arc_points(x_tan, y_tan, x_1, y_1, x_cb, y_cb, x_arc_1[-2], y_arc_1[-2], arc_2_pts, direction='ccw')

    return list(x_arc_1) + list(x_arc_2), list(y_arc_1) + list(y_arc_2), x_ca, y_ca, x_tan, y_tan, start_angle, i_center, x_cb, y_cb, r_1

def get_arc_arc_points(x_0, y_0, x_1, y_1, x_prev, y_prev, r_0, r_1, pts=10, sym_funcs=None, numerical=False):
    
    start_angle = get_line_ori(x_0, y_0, x_prev, y_prev)
    
    x_ca_multi = [x_0 + r_0*np.cos(start_angle + s*np.pi/2) for s in [1, -1]]
    y_ca_multi = [y_0 + r_0*np.sin(start_angle + s*np.pi/2) for s in [1, -1]]

    x_ca = None
    y_ca = None
    i_center = None

    # Get the center point closest to pt 1
    for i_ca, (x, y) in enumerate(zip(x_ca_multi, y_ca_multi)):
        if x_ca is None or norm([x_1-x, y_1-y]) < norm([x_1-x_ca, y_1-y_ca]):
            x_ca = x
            y_ca = y
            i_center = i_ca
    
    x_ca_sym = sp.Symbol('x_ca_sym', real=True)
    y_ca_sym = sp.Symbol('y_ca_sym', real=True)    
    x_cb_sym = sp.Symbol('x_cb_sym', real=True)
    y_cb_sym = sp.Symbol('y_cb_sym', real=True)
    x_1_sym = sp.Symbol('x_1_sym', real=True)
    y_1_sym = sp.Symbol('y_1_sym', real=True)
    x_2_sym = sp.Symbol('x_2_sym', real=True)
    y_2_sym = sp.Symbol('y_2_sym', real=True)
    x_3_sym = sp.Symbol('x_3_sym', real=True)
    y_3_sym = sp.Symbol('y_3_sym', real=True)
    r_a_sym = sp.Symbol('r_a_sym', real=True)
    r_b_sym = sp.Symbol('r_b_sym', real=True)

    if numerical is False:
        if sym_funcs is None and os.path.exists(ARCARC_FILE) is False:
            eq_pt2_on_A = sp.Eq((x_2_sym - x_ca_sym)**2 + (y_2_sym - y_ca_sym)**2, r_a_sym**2)
            eq_pt2_on_B = sp.Eq((x_2_sym - x_cb_sym)**2 + (y_2_sym - y_cb_sym)**2, r_b_sym**2)
            eq_pt3_on_B = sp.Eq((x_3_sym - x_cb_sym)**2 + (y_3_sym - y_cb_sym)**2, r_b_sym**2)
            eq_tangent_at_2 = sp.Eq((x_ca_sym - x_cb_sym)**2 + (y_ca_sym - y_cb_sym)**2, (r_b_sym - r_a_sym)**2)
            
            partial_sols_1 = sp.solve([eq_pt2_on_A, eq_pt2_on_B], x_2_sym, y_2_sym)
            partial_sols_2 = sp.solve([eq_pt3_on_B, eq_tangent_at_2], x_cb_sym, y_cb_sym)

            sym_sols = [[partial_sols_1[idx_sol][idx_var].subs([(x_cb_sym, partial_sols_2[idx_sol][0]), (y_cb_sym, partial_sols_2[idx_sol][1])]) for idx_var in range(2)] + list(partial_sols_2[idx_sol]) for idx_sol in range(2)]

            with open(ARCARC_FILE, 'wb') as fid:
                # dump(sym_sols, fid)
                sym_funcs = lambdify_and_dump(sym_sols, [x_ca_sym, y_ca_sym, x_1_sym, y_1_sym, x_3_sym, y_3_sym, r_a_sym, r_b_sym], fid)

        elif os.path.exists(ARCARC_FILE) is True:
            with open(ARCARC_FILE, 'rb') as fid:
                # sym_sols = load(fid)
                sym_funcs = load(fid)

        # Substitute for symbols in the symbolic solution
        sols = [tuple([f(x_ca, y_ca, x_0, y_0, x_1, y_1, r_0, r_1) for f in sym_func]) for sym_func in sym_funcs]
        
    else:
        eq_pt2_on_A = sp.Eq((x_2_sym - x_ca)**2 + (y_2_sym - y_ca)**2, r_0**2)
        eq_pt2_on_B = sp.Eq((x_2_sym - x_cb_sym)**2 + (y_2_sym - y_cb_sym)**2, r_1**2)
        eq_pt3_on_B = sp.Eq((x_1 - x_cb_sym)**2 + (y_1 - y_cb_sym)**2, r_1**2)
        eq_tangent_at_2 = sp.Eq((x_ca - x_cb_sym)**2 + (y_ca - y_cb_sym)**2, (r_1 - r_0)**2)

        init = (float(np.mean([x_0, x_1])), float(np.mean([y_0, y_1])), x_1, float(np.mean([y_0, y_1])))
        sols = sp.nsolve((eq_pt2_on_A, eq_pt2_on_B, eq_pt3_on_B, eq_tangent_at_2), (x_2_sym, y_2_sym, x_cb_sym, y_cb_sym), init, verify=False)
        sols = [[sols[row, col] for row in range(sols.rows)] for col in range(sols.cols)]

    # Get the best solution  
    best_sol = _get_best_sol2(x_0, y_0, x_1, y_1, x_ca, y_ca, x_prev, y_prev, r_0, r_1, sols)    
    x_tan, y_tan, x_cb, y_cb = map(float, best_sol)
    
    # Figure out how many points in the two arcs (scale by length)
    arc1_length = r_0 * get_0_to_180(get_3point_ang((x_0, y_0), (x_ca, y_ca), (x_tan, y_tan)))
    arc2_length = r_1 * get_0_to_180(get_3point_ang((x_tan, y_tan), (x_cb, y_cb), (x_1, y_1)))
    arc_2_pts = round(pts/(arc1_length/arc2_length + 1))
    arc_1_pts = pts - arc_2_pts if arc_2_pts < pts else 1

    # Generate the points
    x_arc_1, y_arc_1 = get_arc_points(x_0, y_0, x_tan, y_tan, x_ca, y_ca, x_prev,      y_prev,      arc_1_pts, direction='cw')
    x_arc_2, y_arc_2 = get_arc_points(x_tan, y_tan, x_1, y_1, x_cb, y_cb, x_arc_1[-2], y_arc_1[-2], arc_2_pts, direction='cw')

    return list(x_arc_1) + list(x_arc_2), list(y_arc_1) + list(y_arc_2), x_ca, y_ca, x_tan, y_tan, sym_funcs, start_angle, i_center, x_cb, y_cb

def lambdify_and_dump(sols, vbs, fid):
    pk = [[sp.lambdify(vbs, eq, 'numpy') for eq in sol] for sol in sols]
    dill.settings['recurse'] = True
    dill.dump(pk, fid)
    return pk

def subs_lambdify(eq, vbs):
    f = sp.lambdify([vb[0] for vb in vbs], eq, 'numpy')
    ans = f(*[vb[1] for vb in vbs])
    return ans

def drop_zero_imag(sol):
    for i, eq in enumerate(sol):
        if eq.is_Add is True:
            sol[i] = eq.as_real_imag()[0]

    return sol

def get_0_to_180(ang):
    if abs(ang) > np.pi:
        ang = 2*np.pi-abs(ang)
    
    return ang

def get_0_to_360(ang):
    if ang < 0:
        ang = 2*np.pi + ang
    return ang

def get_3point_ang(a, b, c):
    ang = math.atan2(c[1]-b[1], c[0]-b[0]) - math.atan2(a[1]-b[1], a[0]-b[0])

    if ang < 0:
        ang += 2*np.pi
    
    if ang > np.pi:
        ang = 2*np.pi - ang

    return ang

def get_arc_line_points(x_0, y_0, x_1, y_1, x_prev, y_prev, r_curve, pts=10, sym_sols=None, numerical=False):
    
    start_angle = get_line_ori(x_0, y_0, x_prev, y_prev)
    
    x_c_multi = [x_0 + r_curve*np.cos(start_angle + s*np.pi/2) for s in [1, -1]]
    y_c_multi = [y_0 + r_curve*np.sin(start_angle + s*np.pi/2) for s in [1, -1]]

    x_c = None
    y_c = None
    i_center = None

    # Get the center point closest to pt 1
    for i_c, (x, y) in enumerate(zip(x_c_multi, y_c_multi)):
        if x_c is None or norm([x_1-x, y_1-y]) < norm([x_1-x_c, y_1-y_c]):
            x_c = x
            y_c = y
            i_center = i_c
    
    x = sp.Symbol('x', real=True)
    y = sp.Symbol('y', real=True)
    x_c_sym = sp.Symbol('x_c_sym', real=True)
    y_c_sym = sp.Symbol('y_c_sym', real=True)
    x_1_sym = sp.Symbol('x_1_sym', real=True)
    y_1_sym = sp.Symbol('y_1_sym', real=True)
    r_curve_sym = sp.Symbol('r_curve_sym', real=True)

    if numerical is False:
        if sym_sols is None and os.path.exists(ARCLINE_FILE) is False:
            eq_circle = sp.Eq((x - x_c_sym)**2 + (y - y_c_sym)**2, r_curve_sym**2)
            eq_perp = sp.Eq((y - y_c_sym)/(x - x_c_sym), -(x - x_1_sym)/(y - y_1_sym))
            sym_sols = sp.solve([eq_circle, eq_perp], x, y)
            
            with open(ARCLINE_FILE, 'wb') as fid:
                dump(sym_sols, fid)

        elif os.path.exists(ARCLINE_FILE) is True:
            with open(ARCLINE_FILE, 'rb') as fid:
                sym_sols = load(fid)

        # Substitute for symbols in the symbolic solution
        sols = [tuple([eq.subs([(x_c_sym, x_c), (y_c_sym, y_c), (x_1_sym, x_1), (y_1_sym, y_1), (r_curve_sym, r_curve)]) for eq in sym_sol]) for sym_sol in sym_sols]

    else:
        
        eq_circle = sp.Eq((x - x_c)**2 + (y - y_c)**2, r_curve**2)
        eq_perp = sp.Eq((y - y_c)/(x - x_c), -(x - x_1)/(y - y_1))

        init = (50, 150)
        sols = sp.nsolve((eq_circle, eq_perp), (x, y), init, verify=False)
        sols = [[sols[row, col] for row in range(sols.rows)] for col in range(sols.cols)]

    # Get the best solution  
    best_sol = _get_best_sol4(x_0, y_0, x_1, y_1, x_c, y_c, x_prev, y_prev, r_curve, sols)

    x_tan = float(sp.re(best_sol[0]))
    y_tan = float(sp.re(best_sol[1]))
    
    # Figure out how many points in the arc and line (scale by length)
    arc_length = r_curve * get_3point_ang((x_0, y_0), (x_c, y_c), (x_tan, y_tan))
    line_length = norm([x_1 - x_tan, y_1 - y_tan])
    line_pts = round(pts/(arc_length/line_length + 1))
    arc_pts = pts - line_pts if line_pts < pts else 1

    # Generate the points
    x_arc, y_arc = get_arc_points(x_0, y_0, x_tan, y_tan, x_c, y_c, x_prev, y_prev, arc_pts, direction='ccw')
    x_line, y_line = get_line_points(x_tan, y_tan, x_1, y_1, line_pts)

    return list(x_arc) + list(x_line), list(y_arc) + list(y_line), x_c, y_c, x_tan, y_tan, sym_sols, start_angle, i_center


def _get_best_sol3(x_0, y_0, x_1, y_1, x_ca, y_ca, x_prev, y_prev, r_a, tan_end_ang, sols, i_center=1):
    return [sols[row, 0] for row in range(sols.rows)]
    
def _get_best_sol2(x_0, y_0, x_1, y_1, x_ca, y_ca, x_prev, y_prev, r_a, r_b, sols, i_center=1):
    """Gets the solution with the the solution with the shortest arclength
    
    """
    arc_lengths = []
    for i, sol in enumerate(sols):

        # Look for NaNs in tan point and center point
        if any([math.isnan(v) for v in sol[:2]]) and not any([math.isnan(v) for v in sol[2:]]):
            # If NaNs found in tangent point but not in center b, find tangent point from center b            
            x_cb, y_cb = map(float, sol[2:])
            sol = (*get_two_circ_tan_point((x_ca, y_ca), r_a, (x_cb, y_cb), r_b), x_cb, y_cb)
            sols[i] = sol
        
        elif any([math.isnan(v) for v in sol[2:]]):
            # If NaNs found in center point, skip this solution
            arc_lengths.append(np.inf)
            continue
        
        # try:
        x_tan, y_tan, x_cb, y_cb = map(float, sol)
        # except TypeError:
        #     continue

        x_arc_a, y_arc_a = get_arc_points(x_0, y_0, x_tan, y_tan, x_ca, y_ca, x_prev, y_prev)
        x_arc_b, y_arc_b = get_arc_points(x_tan, y_tan, x_1, y_1, x_cb, y_cb, x_arc_a[-2], y_arc_a[-2])

        arc_length_a = sum([norm([x1-x0, y1-y0]) for (x0, x1, y0, y1) in zip(x_arc_a[:-1], x_arc_a[1:], y_arc_a[:-1], y_arc_a[1:])])
        arc_length_b = sum([norm([x1-x0, y1-y0]) for (x0, x1, y0, y1) in zip(x_arc_b[:-1], x_arc_b[1:], y_arc_b[:-1], y_arc_b[1:])])

        arc_lengths.append(sum([arc_length_a, arc_length_b]))
        
    return sols[arc_lengths.index(min(arc_lengths))]

def _get_best_sol(x_0, y_0, x_1, y_1, x_c, y_c, x_prev, y_prev, r_curve, sols, i_center=1):
    """Gets the solution with the the solution with maximum or minimum subtended angle.
    
    """
    angs = []
    for sol in sols:
        
        # Look for NaNs
        if any([math.isnan(v) for v in sol]):
            # If NaNs found, skip this solution
            angs.append(np.inf if i_center==1 else -np.inf)
            continue

        # try:
        x_tan = float(sol[0])
        y_tan = float(sol[1])
        # except TypeError:
        #     continue

        x_arc, y_arc = get_arc_points(x_0, y_0, x_tan, y_tan, x_c, y_c, x_prev, y_prev, 100)

        start_ang = abs(get_3point_ang((x_prev, y_prev), (x_0, y_0), (x_arc[1], y_arc[1])))
        end_ang = abs(get_3point_ang((x_arc[-2], y_arc[-2]), (x_tan, y_tan), (x_1, y_1)))

        angs.append(min([start_ang, end_ang]))
        
    return sols[angs.index(min(angs))] if i_center == 1 else sols[angs.index(max(angs))]

def _get_best_sol4(x_0, y_0, x_1, y_1, x_c, y_c, x_prev, y_prev, r_curve, sols, i_center=1):
    """Gets the solution with the the solution with maximum or minimum subtended angle.
    
    """
    arc_lengths = []
    for sol in sols:
        
        # Look for NaNs
        if any([math.isnan(sp.re(v)) for v in sol]):
            # If NaNs found, skip this solution
            arc_lengths.append(np.inf)
            continue

        x_tan = float(sp.re(sol[0]))
        y_tan = float(sp.re(sol[1]))

        # arc_length = r_curve * get_3point_ang((x_0, y_0), (x_c, y_c), (x_tan, y_tan))
        arc_length = r_curve * get_0_to_360(get_angle_two_vecs(x_0-x_c, y_0-y_c, x_tan-x_c, y_tan-y_c))
        line_length = norm([x_1 - x_tan, y_1 - y_tan])

        arc_lengths.append(arc_length + line_length)
        
    return sols[arc_lengths.index(min(arc_lengths))]

def get_line_points(x_0, y_0, x_1, y_1, pts=10):
    return np.linspace(x_0, x_1, pts), np.linspace(y_0, y_1, pts)

def arc_center(x_0, y_0, x_1, y_1, m_0):
    """Returns the x and y coordinates of the center of curvature of an arc given the start and end coordinates, and the slope at the start.

    Parameters
    ----------
    x_0 : float
        Starting X coordinate
    y_0 : float
        Starting Y coordinate
    x_1 : float
        Ending X coordinate
    y_1 : float
        Ending Y coordinate
    m_0 : float
        Starting slop

    Returns
    -------
    list, list
        x and y coordinates of center of curvature

    """
    a = -m_0
    b = 1

    x = sp.Symbol('x', real=True)
    y = sp.Symbol('y', real=True)

    eq1 = sp.Eq(b*x -a*y - (b*x_0-a*y_0), 0)
    eq2 = sp.Eq((x_0 - x_1)*x + (y_0 - y_1)*y  -  .5*(x_0**2 + y_0**2 - x_1**2 - y_1**2), 0)

    sol = sp.solve([eq1, eq2], x, y)        
    
    x_c = [v for v in sol.values()][0]
    y_c = [v for v in sol.values()][1]

    return x_c, y_c

# def get_arc_points2(x_0, y_0, radius, x_prev, y_prev, delta_angle, pts=10, flip=1, direction=None):

#     entry_angle = get_line_ori(x_0, y_0, x_prev, y_prev)
#     ang_to_center = entry_angle + np.pi/2*flip

#     x_c = x_0 + radius*np.cos(ang_to_center)
#     y_c = y_0 + radius*np.sin(ang_to_center)

#     ang_from_prev = get_line_ori(x_c, y_c, x_0, y_0)  - get_line_ori(x_c, y_c, x_prev, y_prev) 
    
#     delta_angle_1 = delta_angle
#     delta_angle_2 = -np.sign(delta_angle)*(2*np.pi-abs(delta_angle))
    
#     if direction is None:
#         delta_angle = delta_angle_1 if np.sign(delta_angle_1) == np.sign(ang_from_prev) else delta_angle_2
#     elif direction == 'ccw':
#         delta_angle = delta_angle_1 if delta_angle_1 > 0 else delta_angle_2
#     elif direction == 'cw':
#         delta_angle = delta_angle_1 if delta_angle_1 < 0 else delta_angle_2

#     end_angle = start_angle + delta_angle
    
#     angles = np.linspace(start_angle, end_angle, pts)   

#     x = list([x_c + radius*np.cos(angle) for angle in angles])
#     y = list([y_c + radius*np.sin(angle) for angle in angles])

#     return x, y

def get_arc_points(x_0, y_0, x_1, y_1, x_c, y_c, x_prev, y_prev, pts=10, direction=None):
    """Returns points on an arc given the start and end points, center of curvature, and a third point through which the start tangent passes.

    Parameters
    ----------
    x_0 : float
        Starting X coordinate
    y_0 : float
        Starting Y coordinate
    x_1 : float
        Ending X coordinate
    y_1 : float
        Ending Y coordinate
    x_c : float
        X coordinate of center of curvature
    y_c : float
        Y coordinate of center of curvature
    x_prev : float
        X coordinate of point through which the start tangent passes
    y_prev : float
        Y coordinate point through which the start tangent passes
    pts : int, optional
        Number of points to return, by default 10

    Returns
    -------
    list, list
        x and y coordinate of points on the arc

    """
    start_angle = get_line_ori(x_c, y_c, x_0, y_0)

    ang_from_prev = get_line_ori(x_c, y_c, x_0, y_0)  - get_line_ori(x_c, y_c, x_prev, y_prev) 

    # ang_from_prev = get_3point_ang((x_0, y_0), (x_c, y_c), (x_1, y_1))
    
    # delta_angle_1 = get_3point_ang((x_0, y_0), (x_c, y_c), (x_1, y_1))    
    delta_angle_1 = get_line_ori(x_c, y_c, x_1, y_1)  - get_line_ori(x_c, y_c, x_0, y_0) 
    delta_angle_2 = -np.sign(delta_angle_1)*(2*np.pi-abs(delta_angle_1))
    
    if direction is None:
        delta_angle = delta_angle_1 if np.sign(delta_angle_1) == np.sign(ang_from_prev) else delta_angle_2
    elif direction == 'ccw':
        delta_angle = delta_angle_1 if delta_angle_1 > 0 else delta_angle_2
    elif direction == 'cw':
        delta_angle = delta_angle_1 if delta_angle_1 < 0 else delta_angle_2

    end_angle = start_angle + delta_angle
    
    angles = np.linspace(start_angle, end_angle, pts)   
    radius = norm(np.array([x_0-x_c, y_0-y_c]).astype(np.float64))

    x = list([x_c + radius*np.cos(angle) for angle in angles])
    y = list([y_c + radius*np.sin(angle) for angle in angles])

    return x, y

def get_two_circ_tan_point(loc_ca, r_a, loc_cb, r_b):
    def error(loc):
        err_a = abs(norm([loc[0] - loc_ca[0], loc[1] - loc_ca[1]]) - r_a)
        err_b = abs(norm([loc[0] - loc_cb[0], loc[1] - loc_cb[1]]) - r_b)
        return max([err_a, err_b])
    
    res = optimize.minimize(error, [-10, 10], method='Nelder-Mead')
    return tuple(res.x)

def get_line_ori(x_0, y_0, x_1, y_1):
    """Returns the orientation of a vector defined by two points in 2D space.

    Parameters
    ----------
    x_0 : float
        X coordinate of first point
    y_0 : float
        Y coordinate of first point
    x_1 : float
        X coordinate of second point
    y_1 : float
        Y coordinate of second point

    Returns
    -------
    float
        Orientation of the vector (radians)

    """
    
    return math.atan2(y_1-y_0, x_1-x_0)

def get_angle_two_vecs(x_0, y_0, x_1, y_1):
    """Returns the orientation between the two vectors defined from the origin to the two given points in 2D space.

    Parameters
    ----------
    x_0 : float
        X coordinate of first point
    y_0 : float
        Y coordinate of first point
    x_1 : float
        X coordinate of second point
    y_1 : float
        Y coordinate of second point

    Returns
    -------
    float
        Angle between the two vectors (radians)

    """
    ang_0 = get_line_ori(0, 0, x_0, y_0)
    ang_1 = get_line_ori(0, 0, x_1, y_1)
    ang = ang_1 - ang_0

    return ang


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    x_0 = -66.411
    y_0 = 57.325
    x_1 = -59.795
    y_1 = 117.916
    r_0 = 27
    r_1 = 60
    x_prev = -53.044
    y_prev = 55.998

    x_arc, y_arc, x_ca, y_ca, x_tan, y_tan, sym_sols, start_angle, i_center, x_cb, y_cb = get_arc_arc_points(x_0, y_0, x_1, y_1, x_prev, y_prev, r_0, r_1, pts=10)

    (fig, ax) = plt.subplots()
    ax.scatter(x_arc, y_arc)
    ax.scatter([x_0, x_1], [y_0, y_1], c='r')
    ax.scatter([x_cb, x_ca], [y_cb, y_ca], c='g')
    ax.grid()
    ax.axis('square')

    plt.show()


import os
import unittest

from numpy import genfromtxt
import matplotlib.pyplot as plt
from thornpy.signal import manually_clean_sig, manually_clean_sigs

TEST_SPIKE_DATA = os.path.join('test', 'files', 'spike.csv')

class Test_ManuallyCleanSig(unittest.TestCase):

    def setUp(self):
        return

    def test_manually_clean_sig_with_y(self):
        data = genfromtxt(TEST_SPIKE_DATA, delimiter=',', names=True)
        
        x = data['time']
        y = data['stress']

        y = manually_clean_sig(x, y, print_debug=True)

        _, ax = plt.subplots()
        
        ax.plot(x, y)
        ax.grid()
        ax.set_title('Cleaned Signal')
        plt.show()

        self.assertEqual(y[1443], -176000.0)

    def test_manually_clean_sig_return_indices(self):
        data = genfromtxt(TEST_SPIKE_DATA, delimiter=',', names=True)
        
        x = data['time']
        y = data['stress']

        y, i = manually_clean_sig(x, y, print_debug=True, indices=True)

        _, ax = plt.subplots()
        
        ax.plot(x, y)
        ax.grid()
        ax.set_title('Cleaned Signal')
        plt.show()

        self.assertListEqual(i, [1443])

    def test_manually_clean_multiple_sigs(self):
        data = genfromtxt(TEST_SPIKE_DATA, delimiter=',', names=True)
        
        x = data['time']
        y = [data['stress'], [v/2 for v in data['stress']]]

        y = manually_clean_sigs(x, y, print_debug=True)

        _, ax = plt.subplots()
        
        for each_y in y:
            ax.plot(x, each_y)
        ax.grid()
        ax.set_title('Cleaned Signal')
        plt.show()

        self.assertEqual(y[1][1443], -176000.0/2)

    def tearDown(self):
        return


import os
import unittest

from pandas import read_csv
import matplotlib.pylab as plt

from thornpy.signal import fft_watefall

DATA_FILENAME = os.path.join(os.getcwd(), 'test', 'files', 'waterfall', 'LC3_aluminum_pt1_accel_X.csv')

class Test_Waterfall(unittest.TestCase):

    # def setUp(self):        
    #     self.wtrfl, self.time, self.rpm, self.sig, self.order_cut_plots = self._make_plots('linear')

    def test_order_plots_human_approval(self):
        plt.show()
        self.assertTrue(True)

    def test_db(self):
        _wtrfl_db, _time_db, _rpm_db, _sig_db, _order_cut_plots_db = self._make_plots('dB')
        plt.show()
        self.assertTrue(False)

    def test_db_vmin_vmax(self):
        vmin = -100
        vmax = -5  
        _wtrfl_db, _time_db, _rpm_db, _sig_db, _order_cut_plots_db = self._make_plots('dB', vmin=vmin, vmax=vmax)
        plt.show()
        self.assertTrue(False)

    def tearDown(self):
        return

    def _make_plots(self, z_scale, vmin=None, vmax=None):
        time, rpm, response = _get_data()
        return fft_watefall(time, response, percent_overlap=75, n_fft=1024, t_min=6.25, t_max=10, z_scale=z_scale, f_range=[0, 500], return_order_cuts=[1, 40, 63, 80], order_lines=[1, 40, 63, 80], input_sig=rpm, input_conversion_factor=1, response_unit='g', vmin=vmin, vmax=vmax)

def _get_data():
    data = read_csv(DATA_FILENAME)
    
    time = list(data['Time (s)'])
    rpm = list(data['Shaft Speed (rpm)'])
    response = list(data['Response (g)'])

    return time, rpm, response


from pathlib import Path
from tempfile import TemporaryDirectory
import unittest
from thornpy.plotting import plot_3d, save_multifig
import matplotlib.pyplot as plt
import numpy as np

class Test_Plotting(unittest.TestCase):

    def setUp(self):
        return

    def test_plot_3d(self):
        x = [1,2,3,4,5]
        y = [10, 15, 20, 15, 10]
        z = [0, 0, 0, 0, 0]
        figure = plot_3d(x, y, z)
        plt.show()
        
        self.assertEqual(0,1)

    def test_concat_figs(self):
        x = np.array([1,2,3,4,5])
        y = np.array([10, 15, 20, 15, 10])
        fig_1, ax_1 = plt.subplots()
        ax_1.plot(x, y)
        fig_2, ax_2 = plt.subplots()
        ax_2.plot(x, y*2)

        with TemporaryDirectory() as tmp:
            png_file = Path(tmp) / 'tmp.png'
            fig = save_multifig([fig_1, fig_2], png_file)
            self.assertTrue(png_file.exists())
    
    def tearDown(self):
        return

    


"""Tests the functions in the numtype module.
"""
import unittest
from thornpy import numtype

FLOAT_STRING = '1.0'
NEG_FLOAT_STRING = '-1.0'
INT_STRING = '1'
NEG_INT_STRING = '-1'
CHAR_STRING = 'lskfsd'
EMPTY_STRING = ''

class Test_StrIsFloat(unittest.TestCase):
    """Tests that numtype.str_is_float() returns correct booleans
    """
    def setUp(self):
        return

    def test_float(self):
        """Test that str_is_float() returns true when a float is passed
        """
        self.assertTrue(numtype.str_is_float(FLOAT_STRING))

    def test_int(self):
        """Test that str_is_float() returns false when an int is passed
        """
        self.assertFalse(numtype.str_is_float(INT_STRING))

    def tearDown(self):
        return

class Test_StrIsInt(unittest.TestCase):
    """Tests that numtype.str_is_int() returns correct booleans
    """

    def setUp(self):
        return

    def test_int(self):
        """Test that str_is_int() returns true when an int is passed
        """
        self.assertTrue(numtype.str_is_int(INT_STRING))

    def test_float(self):
        """Test that str_is_int() returns true when an int is passed
        """
        self.assertFalse(numtype.str_is_int(FLOAT_STRING))

    def tearDown(self):
        return

class Test_StrIsPosNum(unittest.TestCase):
    """Tests that numtype.str_is_pos_num() returns correct booleans
    """

    def setUp(self):
        return

    def test_int(self):
        """Test that str_is_int() returns true when an int is passed
        """
        self.assertTrue(numtype.str_is_pos_num(INT_STRING))

    def test_float(self):
        """Test that str_is_int() returns true when a float is passed
        """
        self.assertTrue(numtype.str_is_pos_num(FLOAT_STRING))

    def test_neg_int(self):
        """Test that str_is_int() returns true when a negative int is passed
        """
        self.assertFalse(numtype.str_is_pos_num(NEG_INT_STRING))

    def test_neg_float(self):
        """Test that str_is_int() returns true when a negative float is passed
        """
        self.assertFalse(numtype.str_is_pos_num(NEG_FLOAT_STRING))

    def tearDown(self):
        return

class Test_StrIsNum(unittest.TestCase):
    """Tests that numtype.str_is_pos_num() returns correct booleans
    """

    def setUp(self):
        return

    def test_int(self):
        """Test that str_is_int() returns true when an int is passed
        """
        self.assertTrue(numtype.str_is_num(INT_STRING))

    def test_float(self):
        """Test that str_is_int() returns true when a float is passed
        """
        self.assertTrue(numtype.str_is_num(FLOAT_STRING))

    def test_neg_int(self):
        """Test that str_is_int() returns true when a negative int is passed
        """
        self.assertTrue(numtype.str_is_num(NEG_INT_STRING))

    def test_neg_float(self):
        """Test that str_is_int() returns true when a negative float is passed
        """
        self.assertTrue(numtype.str_is_num(NEG_FLOAT_STRING))

    def tearDown(self):
        return
